// CatGame.java

import ch.aplu.jgamegrid.*;
import ch.aplu.util.*;
import java.util.*;
import java.awt.event.KeyEvent;
import java.awt.*;

public class CatGame extends GameGrid
{
  private final int nbHorzTiles = 30;
  private final int nbVertTiles = 5;
  private final int tileSize = 64;
  private Dog[] dogs = new Dog[2];
  private Cat cat;
  GGTileMap tm;
  private boolean isFliesEnabled = false;
  private int nbEatenFlies = 0;
  private final String info =
    "Welcome to the Cat Game!\n\n" +
    "The cat wants to move to right where it finds delicious\n" +
    "flies to eat. The dogs are dangerous, so the cat\n" +
    "should keep away from their heads.\n\n" +
    "Use the left and right cursor keys to move the cat,\n" +
    "the space key will make the cat jump.\n\n" +
    "There is only a limited number of flies and the cat\n" +
    "must catch them with his mouth.\n\n" +
    "The aim is to eat as much flies as possible\n" +
    "(the number of eaten flies is shown in the title bar).\n\n" +
    "If a dog catches you, ESC brings you back to the start.\n" +
    "Good luck!";

  public CatGame()
  {
    super(640, 320, 1, null, "sprites/clouds.gif", false);
    setTitle("CatGame - www.aplu.ch/jgamegrid");
    new ModelessOptionPane(20, 20, info, null);
    setSimulationPeriod(50);

    tm = createTileMap(nbHorzTiles, nbVertTiles, tileSize, tileSize);
    for (int i = 0; i < nbHorzTiles; i++)
      tm.setImage("sprites/tile_B.gif", i, 4);

    // first bumb at i = 2, 3
    tm.setImage("sprites/tile_G.gif", 2, 4);
    tm.setImage("sprites/tile_H.gif", 3, 4);

    tm.setImage("sprites/tile_E.gif", 2, 3);
    tm.setImage("sprites/tile_F.gif", 3, 3);

    // second bumb at i = 10, 11
    tm.setImage("sprites/tile_G.gif", 10, 4);
    tm.setImage("sprites/tile_H.gif", 11, 4);

    tm.setImage("sprites/tile_C.gif", 10, 3);
    tm.setImage("sprites/tile_D.gif", 11, 3);

    tm.setImage("sprites/tile_E.gif", 10, 2);
    tm.setImage("sprites/tile_F.gif", 11, 2);

    // third bumb at i = 20, 21, 22, 23
    tm.setImage("sprites/tile_G.gif", 20, 4);
    tm.setImage("sprites/tile_A.gif", 21, 4);
    tm.setImage("sprites/tile_A.gif", 22, 4);
    tm.setImage("sprites/tile_H.gif", 23, 4);

    tm.setImage("sprites/tile_E.gif", 20, 3);
    tm.setImage("sprites/tile_G.gif", 21, 3);
    tm.setImage("sprites/tile_H.gif", 22, 3);
    tm.setImage("sprites/tile_F.gif", 23, 3);

    tm.setImage("sprites/tile_E.gif", 21, 2);
    tm.setImage("sprites/tile_F.gif", 22, 2);

    dogs[0] = new Dog(this);
    dogs[0].addCollisionTile(new Location(3, 3));
    dogs[0].addCollisionTile(new Location(10, 3));
    dogs[0].setCollisionCircle(0, new Point(22, -25), 10);
    dogs[0].setCollisionCircle(1, new Point(22, -25), 10);
    addActor(dogs[0], new Location(350, 225));

    dogs[1] = new Dog(this);
    dogs[1].addCollisionTile(new Location(11, 3));
    dogs[1].addCollisionTile(new Location(20, 3));
    dogs[1].setCollisionCircle(0, new Point(28, -16), 15);
    dogs[1].setCollisionCircle(1, new Point(28, -16), 15);
    addActor(dogs[1], new Location(850, 220));

    cat = new Cat(this);
    addActor(cat, new Location(30, 235));
    addKeyListener(cat);
    cat.setCollisionRectangle(new Point(0, 0), 35, 55);

    ArrayList<Location> tiles = new ArrayList<Location>();
    // First bumb
    tiles.add(new Location(2, 3));
    tiles.add(new Location(3, 3));

    // Second bumb
    tiles.add(new Location(10, 3));
    tiles.add(new Location(10, 2));
    tiles.add(new Location(11, 3));
    tiles.add(new Location(11, 2));

    // Third bumb
    tiles.add(new Location(20, 3));
    tiles.add(new Location(21, 2));
    tiles.add(new Location(22, 2));
    tiles.add(new Location(23, 3));

    // Comment out these lines for development tests
    cat.addCollisionTiles(tiles);
    cat.addCollisionActor(dogs[0]);
    cat.addCollisionActor(dogs[1]);

    show();
    doRun();
    Monitor.putSleep();
    int nbFlies = 0;
    while (nbFlies < 50)
    {
      Location randomLocation =
        new Location((int)(400 + 200 * Math.random()), 250);
      Fly fly = new Fly(this);
      addActor(fly, randomLocation);
      fly.addCollisionActor(cat);
      delay((long)(200 + 1000 * Math.random()));
      nbFlies++;
    }
  }

  public void reset()
  {
    tm.setPosition(new Point(0, 0));
    setBgImagePos(new Point(0, 0));
  }

  public void setFliesEnabled(boolean enabled)
  {
    isFliesEnabled = true;
    Monitor.wakeUp();
    cat.setCollisionCircle(new Point(16, -15), 5);
  }

  public boolean isFliesEnabled()
  {
    return isFliesEnabled;
  }

  public void increaseEatenFlies()
  {
    nbEatenFlies++;
    setTitle("Flies eaten: " + nbEatenFlies);
  }

  public Dog[] getDogs()
  {
    return dogs;
  }

  public Cat getCat()
  {
    return cat;
  }

  public static void main(String[] args)
  {
    new CatGame();
  }
}


// --------------------- class Cat ----------------------------------------
class Cat extends Actor
  implements GGKeyListener, GGTileCollisionListener, GGActorCollisionListener
{
  private Location location;
  private boolean downward;
  private boolean forward;
  private boolean backward;
  private boolean onTop;
  private CatGame cg;
  private int tmX;
  private int tmY;
  private boolean isGameOver;

  public Cat(CatGame cg)
  {
    super("sprites/cat.gif", 3);
    this.cg = cg;
    addTileCollisionListener(this);
  }

  public void reset()
  {
    tmX = 0;
    tmY = 0;
    downward = false;
    forward = false;
    backward = false;
    onTop = false;
    isGameOver = false;
    setHorzMirror(false);
    location = getLocationStart();
  }

  public void act()
  {
    // -------- Move cat ----------
    if (downward && !onTop)
    {
      location.y += 5;
      if (location.y == 235)
        downward = false;
      setLocation(location);
    }
    if (forward)
    {
      forward = false;
      location.x += 10;
      if (!downward || onTop)
        show((getIdVisible() == 0 ? 1 : 0));
      setLocation(location);
    }
    if (backward)
    {
      backward = false;
      if (!downward || onTop)
        show((getIdVisible() == 0 ? 1 : 0));
      location.x -= 10;
      if (cg.isFliesEnabled())
        location.x = Math.max(380, location.x);  // inhibit to go back
      setLocation(location);
    }
    onTop = false;

    // -------- Move tile map and background image ------------
    if (cg.isFliesEnabled())  // Inhibit movement of map
      return;
    GGTileMap tm = cg.getTileMap();
    Dog[] dogs = cg.getDogs();
    location = getLocation();
    if (location.x > 400)
    {
      // Move tile map
      tmX -= 5;
      if (tmX < -1280)
        tmX = -1280;
      else
      {
        location.x = 400;
        setLocation(location);
        tm.setPosition(new Point(tmX, tmY));
        cg.setBgImagePos(new Point(tmX, 0));
        dogs[0].setX(dogs[0].getX() - 5);
        dogs[1].setX(dogs[1].getX() - 5);
      }

    }
    if (location.x < 300)
    {
      // Move tile map
      tmX += 5;
      if (tmX > 0)
        tmX = 0;
      else
      {
        location.x = 300;
        setLocation(location);
        tm.setPosition(new Point(tmX, tmY));
        gameGrid.setBgImagePos(new Point(tmX, 0));
        dogs[0].setX(dogs[0].getX() + 5);
        dogs[1].setX(dogs[1].getX() + 5);
      }
    }
    if (tmX < -1180)
      cg.setFliesEnabled(true);

  }

  public boolean keyPressed(KeyEvent evt)
  {
    switch (evt.getKeyCode())
    {
      case KeyEvent.VK_SPACE:
        if (downward && !onTop)
          return true;
        location = getLocation();
        location.y = 50;
        setLocation(location);
        downward = true;
        break;
      case KeyEvent.VK_LEFT:
        location = getLocation();
        setHorzMirror(true);
        backward = true;
        break;
      case KeyEvent.VK_RIGHT:
        location = getLocation();
        if (location.x > 600)
          return true;
        setHorzMirror(false);
        forward = true;
        break;
      case KeyEvent.VK_ESCAPE:
        if (isGameOver)
        {
          cg.doReset();
          cg.doRun();
          cg.setTitle("CatGame - www.aplu.ch/jgamegrid");
        }
        break;
    }
    return true;
  }

  public boolean keyReleased(KeyEvent evt)
  {
    return true;
  }

  // Assumed to be called after act()
  public int collide(Actor actor, Location location)
  {
    int a = 10;  // distance to jump back
    // First bumb
    if (location.equals(new Location(2, 3)))
      walk(-a, 170);
    if (location.equals(new Location(3, 3)))
      walk(a, 170);

    // Second bumb
    if (location.equals(new Location(10, 3)) ||
      location.equals(new Location(10, 2)))
      walk(-a, 110);
    if (location.equals(new Location(11, 3)) ||
      location.equals(new Location(11, 2)))
      walk(a, 110);

    // Third bumb lower part
    if (location.equals(new Location(20, 3)))
      walk(-a, 170);
    if (location.equals(new Location(23, 3)))
      walk(a, 170);

    // Third bumb upper part
    if (location.equals(new Location(21, 2)))
      walk(-a, 110);
    if (location.equals(new Location(22, 2)))
      walk(a, 110);

    return 0;
  }

  private void walk(int distance, int height)
  {
    if (getLocation().y > height)
      setX(getX() + distance);
    else
      onTop = true;
  }

  public int collide(Actor a1, Actor a2)
  {
    a1.hide();
    a2.show(2);
    cg.doPause();  // The current simulation cycle is run to the end
    cg.setTitle("Game over. Press ESC to restart.");
    isGameOver = true;
    return 0;
  }
}


// ------------------------- class Dog --------------------------------------
class Dog extends Actor
{
  private CatGame cg;
  private int skipAct;

  public Dog(CatGame cg)
  {
    super("sprites/dog.gif", 3);
    this.cg = cg;
  }

  public void reset()
  {
    skipAct = 0;
    setHorzMirror(false);
  }

  public void act()
  {
    move(3);
    if (getIdVisible() == 1)
      show(0);
    else
      if (getIdVisible() == 0)
        show(1);
    if (skipAct > 0)
    {
      skipAct--;
      return;
    }
    // This will make the game more difficult:
    // If dog sees the cat on the base line, he turns toward the cat
    int dist = getX() - cg.getCat().getX();
    if (cg.getCat().getY() == 235)
    {
      if (dist > 0)
        turnLeft();
      else
        turnRight();
    }
  }

  public int collide(Actor actor, Location location)
  {
    if (location.equals(new Location(3, 3)) ||
      location.equals(new Location(11, 3)))
      turnRight();
    if (location.equals(new Location(10, 3)) ||
      location.equals(new Location(20, 3)))
     turnLeft();
    skipAct = 80;
    return 0;
  }

  private void turnRight()
  {
    setDirection(0);
    setHorzMirror(false);
  }

  private void turnLeft()
  {
    setDirection(180);
    setHorzMirror(true);
  }

}


// ----------------------- class Fly --------------------------------------
class Fly extends Actor
{
  private CatGame cg;

  public Fly(CatGame cg)
  {
    super(true, "sprites/fly1.gif");
    this.cg = cg;
  }

  public void act()
  {
    setDirection(225 + 90 * Math.random());
    move();
    if (getY() < 10)
      removeSelf();
  }

  public int collide(Actor a1, final Actor a2)
  {
    cg.increaseEatenFlies();
    a2.show(2);
    cg.refresh();
    new Thread()
    {
      public void run()
      {
        delay(400);
        a2.show(0);
      }
    }.start();
    removeSelf();
    return 0;
  }
}
